package org.insa.graphs.algorithm.shortestpath;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import org.insa.graphs.algorithm.AbstractSolution.Status;
import org.insa.graphs.algorithm.utils.BinaryHeap;
import org.insa.graphs.model.Arc;
import org.insa.graphs.model.Graph;
import org.insa.graphs.model.Node;
import org.insa.graphs.model.Path;

public class DijkstraAlgorithm extends ShortestPathAlgorithm {

    public DijkstraAlgorithm(ShortestPathData data) {
        super(data);
    }

    @Override
    protected ShortestPathSolution doRun() {
        final ShortestPathData data = getInputData();
        ShortestPathSolution solution = null;
        // TODO:

        // Heap of "sommets"
        BinaryHeap<Label> labels = new BinaryHeap<Label>(); // change to a comparable type

        Graph graph = data.getGraph();
        final int nbNodes = graph.size();

        // Initialize array of distances.
        double[] distances = new double[nbNodes];
        Arrays.fill(distances, Double.POSITIVE_INFINITY);
        distances[data.getOrigin().getId()] = 0;

        // Notify observers about the first event (origin processed).
        notifyOriginProcessed(data.getOrigin());

        // Init Dijkstra
        //ArrayList<Label> labels = new ArrayList<Label>();
        for (Node node : graph.getNodes()) {
            labels.insert(new Label(node));
        }
        // Nodes set
        for (Label label : labels.array) {
            label.marked = false;
            label.setPathCost(Float.MAX_VALUE);
            label.parentNode = null;
        }
        // Origin set
        Label s = labels.array.get(0);
        s.setPathCost(0);
        labels.insert(s);

        boolean trouve = false; // TODO : breaking condition
        Label x;
        while (!labels.isEmpty() && !trouve) {
            x = labels.deleteMin();
            x.mark();

            // y shall be all the successors of x
            float arc_cost = 0;
            for (Label y : labels.array) {
                if (y.getParentNode() == x.getNode()) {
                    if (y.isMarked()) {
                        // This loop serves to get the lentgh of the arc as 
                        // we know its origin and destination
                        for (Arc arc : x.getNode().getSuccessors()) {
                            if (y.getNode() == arc.getDestination()) {
                                arc_cost = arc.getLength();
                            }
                        }
                        float new_cost = x.getCost() + arc_cost;
                        if (y.getCost() > new_cost) {
                        y.setPathCost(new_cost);
                        labels.insert(y);
                        y.setParentNode(x.getNode());
                        }
                    }
            }
            }
        }

        // Create the path ...
        ArrayList<Arc> arcs_path = new ArrayList<>();
        Label dest = labels.findMin();
        for (Label label : labels.array) {
            if (label.getNode() == data.getDestination()){
                dest = label;
            }
        }

        Node current = dest.getParentNode();
        while (current != null) {
            for (Arc arc : current.getSuccessors()) {
                if (arc.getDestination() == dest.getNode()) {
                    arcs_path.add(arc);
                }
            }
            for (Label label : labels.array) {
                if (label.getNode() == current){
                    dest = label;
                    current = label.getParentNode();
                }
            }
        }
        
        // Reverse the path...
        Collections.reverse(arcs_path);

        // Create the final solution.
        solution = new ShortestPathSolution(data, Status.OPTIMAL, new Path(graph, arcs_path));

        return solution;
    }

}
